/*
 * xbwt.hpp for trickier xbwt tricks
 * Copyright (c) 2018 Stefan Stauß, Uwe Baier All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef __XBWT_H
#define __XBWT_H

#include <algorithm>
#include <deque>
#include <sdsl/bit_vectors.hpp>
#include <sdsl/bits.hpp>
#include <sdsl/bp_support.hpp>
#include <sdsl/int_vector.hpp>
#include <sdsl/rank_support.hpp>
#include <sdsl/sdsl_concepts.hpp>
#include <sdsl/select_support.hpp>
#include <sdsl/util.hpp>
#include <sdsl/wavelet_trees.hpp>
#include <stdio.h>

#include "xbwt_construct_algorithm.hpp"

struct xbwt_tag {};

//! xbwt main class
template<class t_wt_type =       sdsl::wt_blcd<>,
         class t_bv_type =       typename t_wt_type::bit_vector_type,
         class t_rank_type =     typename t_bv_type::rank_1_type,
         class t_select_type =   typename t_bv_type::select_1_type,
         class t_bp_support =    sdsl::bp_support_sada<>>
class xbwt {
public:
	typedef xbwt_tag                                index_category;
	typedef sdsl::byte_alphabet_tag                 alphabet_category;
	typedef sdsl::int_vector<>::size_type           size_type;

	typedef sdsl::int_vector<8>                     text_type;

	typedef t_wt_type                               wt_type;
        typedef t_bv_type                               bit_vector_type;
        typedef t_rank_type                             rank_type;
	typedef t_select_type                           select_type;
	typedef t_bp_support                            bp_support; 

	//! delimiting character between strings
	static constexpr unsigned char delimchar = '\1';

	//! used construction algorithm
	static xbwt_construct_algorithm construct_algorithm;

private:
	template<xbwt_construct_algorithm> friend class xbwt_construct;

	wt_type                                         L;

	bit_vector_type                                 Last;
	rank_type                                       Last_rank;
	select_type                                     Last_select;

	sdsl::int_vector<64>                            C;

	sdsl::bit_vector                                P;
	bp_support                                      P_bps_support;

public:
	xbwt() {};

	//! returns id of root node
	size_type root() const {
		return 0;
	};

	//! returns a range of edge identifiers all outgoing edges are contained.
	//! number of childs can be computed using rb - lb (rb is exclusive)
	void get_child_range(size_type i, size_type &lb, size_type &rb) const {
	  lb = (i == 0) ? 0 : Last_select(i) + 1;
	  rb = Last_select(i + 1) + 1;
	};

	//! returns the label of the edge indicated by e.
	//! Combine get_child_range and this function to enumerate all edge labels
	unsigned char get_edge_label(size_type r) const {
	  return (unsigned char)L[r];
	};

	//! returns the id of the child of node i reachable using edge
	//! labeled 'c'. If c != delimchar, returns id of root if no
	//! such child exists. If c == delimchar, returns root() if i is no leaf,
	//! get_rev_lex_rank(i) if i is a leaf (different from root()).
	size_type get_child(size_type i, unsigned char c) const {

	  size_type lb, rb;
	  get_child_range( i, lb, rb );

	  size_type r_lb = L.rank(lb, c);
	  size_type r_rb = L.rank(rb, c);

	  return (r_lb == r_rb) ? root() : C[c] + r_lb;
	};

	//! gets the parent edge label of node i
	unsigned char get_parent_edge(size_type i) const {

	  auto ub = std::upper_bound(C.begin(), C.end(), i);
	  return ub - C.begin() - 1;
	};

	//! returns the id of the parent of the current node
	size_type get_parent(size_type i) const {

	  unsigned char c = get_parent_edge(i);
	  size_type pos = L.select(1 + i - C[c], c);
	  return Last_rank(pos);
	};

	//! returns true if current node is a leaf
	bool is_leaf(size_type i) const {

	  return get_child(i, delimchar) != root();
	};

	//! returns the rank of the lexicographically reversed string
	//! associated to leaf i, starting from 1
	size_type get_rev_lex_rank(size_type i) const {
	  size_type pos = (i == 0) ? 0 : Last_select(i) + 1;
	  return L.rank(pos, delimchar) + 1;
	};

	//! returns the node to which the suffix link from n ode i points to
	//! i mustn't be the root node
	size_type suffix_link(size_type i) const {
	  size_type k = P_bps_support.select(i + 1);
	  size_type j = P_bps_support.enclose(k);
	  size_type l = P_bps_support.rank(j);

	  return l - 1;
	};

	//! compares if two xbw tries were built from the same source
	bool operator==(const xbwt &x) const {
		return L.size() == x.L.size() &&
		       std::equal(L.begin(), L.end(), x.L.begin()) &&
		       Last == x.Last &&
		       C == x.C &&
		       P == x.P;
	};

	//! serializes opbject
	size_type serialize(std::ostream &out, sdsl::structure_tree_node *v,
                          std::string name) const {

		sdsl::structure_tree_node *child =
			sdsl::structure_tree::add_child(v, name, sdsl::util::class_name(*this));
		size_type written_bytes = 0;
		written_bytes += L.serialize(out, child, "L");
		written_bytes += Last.serialize(out, child, "Last");
		written_bytes += Last_rank.serialize(out, child, "Last_rank");
		written_bytes += Last_select.serialize(out, child, "Last_select");
		written_bytes += C.serialize(out, child, "C");
		written_bytes += P.serialize(out, child, "P");
		written_bytes += P_bps_support.serialize(out, child, "P_bps_support");
		sdsl::structure_tree::add_size(child, written_bytes);
		return written_bytes;
	};

	//! loads a serialized object
	void load(std::istream &in) {
		L.load(in);
		Last.load(in);
		Last_rank.load(in, &Last);
		Last_select.load(in, &Last);
		C.load(in);
		P.load(in);
		P_bps_support.load(in, &P);
	};
};

//defaults for construction
template<class wt, class bvt, class rt, class st, class bp_s>
xbwt_construct_algorithm xbwt<wt,bvt,rt,st,bp_s>::construct_algorithm = MAN;

//// SPECIAL CONSTRUCTION FOR XBWT ////////////////////////////////////////////

template <class t_index>
void construct(t_index &idx, const std::string &file, sdsl::cache_config &config,
               uint8_t num_bytes, xbwt_tag) {
  const char *KEY_TEXT =
      sdsl::key_text_trait<t_index::alphabet_category::WIDTH>::KEY_TEXT;
  const char *KEY_BWT =
      sdsl::key_bwt_trait<t_index::alphabet_category::WIDTH>::KEY_BWT;
  typedef sdsl::int_vector<t_index::alphabet_category::WIDTH> text_type;
  {
    // get text
    auto event = sdsl::memory_monitor::event("parse input text");
    text_type text;
    sdsl::load_vector_from_file(text, file, num_bytes);
    if (sdsl::contains_no_zero_symbol(text, file)) {
      sdsl::append_zero_symbol(text);
    }
    sdsl::store_to_cache(text, KEY_TEXT, config);
  }
  {
    // construct suffix array
    auto event = sdsl::memory_monitor::event("SA");
    sdsl::construct_sa<t_index::alphabet_category::WIDTH>(config);
    sdsl::register_cache_file(sdsl::conf::KEY_SA, config);
  }
  {
    // construct BWT
    auto event = sdsl::memory_monitor::event("BWT");
    sdsl::construct_bwt<t_index::alphabet_category::WIDTH>(config);
    sdsl::register_cache_file(KEY_BWT, config);
  }
  {
    // construct XBWT
    auto event = sdsl::memory_monitor::event("XBWT");
    switch (t_index::construct_algorithm) {
    case MAN:
         xbwt_construct<MAN>::construct(config, idx);
         break;
    case MAN_LW:
         xbwt_construct<MAN_LW>::construct(config, idx);
         break;
    case OHL:
         xbwt_construct<OHL>::construct(config, idx);
         break;
    case OHL_LW:
         xbwt_construct<OHL_LW>::construct(config, idx);
         break;
    }
  }
  {
    auto event = sdsl::memory_monitor::event("delete temporary files");
    sdsl::util::delete_all_files(config.file_map);
  }
};

#endif
