/*
 * xbwt_construct_algorithm.hpp for trickier xbwt tricks
 * Copyright (c) 2018 Stefan Stauß, Uwe Baier All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef __XBWT_CONSTRUCT_ALGORITHM_HPP
#define __XBWT_CONSTRUCT_ALGORITHM_HPP

#include <algorithm>
#include <stack>
#include <utility>
#include <vector>
#include <sdsl/int_vector.hpp>
#include <sdsl/construct.hpp>
#include <sdsl/construct_config.hpp>
#include <sdsl/rank_support.hpp>
#include <sdsl/util.hpp>

//! available xbwt construction algorithms
enum xbwt_construct_algorithm {
	MAN,	//Manzini
	MAN_LW,	//Manzini lightweight
	OHL,	//Ohlebusch
	OHL_LW	//Ohlebusch lightweight
};

template<xbwt_construct_algorithm>
class xbwt_construct {
private:
	typedef sdsl::int_vector<>::size_type           size_type;
	typedef sdsl::int_vector<8>                     text_type;
	typedef sdsl::int_vector_buffer<8>              text_buf_type;

	// keys for cache config
	static const char KEY_MR[];
	static const char KEY_RCP[];
	static const char KEY_L[];

public:
	//! main construction method, requires text, bwt and suffix array to reside in cache
	template<class xbwt_t>
	static void construct( sdsl::cache_config &config, xbwt_t &x );

private:
	//! utility function, computes length of longest string (return value) and the C-array
	template<class xbwt_t>
	static size_type computeLmax( sdsl::cache_config &config ) {
		text_buf_type text_buf( sdsl::cache_file_name( sdsl::conf::KEY_TEXT, config ) );
		size_type l_max = 0;
		
		for (size_type i = 0, l = 0; i < text_buf.size() - 1; i++) { //ignore nullbyte at end
			unsigned char c = text_buf[i];
			if (c == xbwt_t::delimchar || c == 0) { //check length of longest string
				l_max = std::max(l_max, l);
				l = 0;
			} else {
				++l;
			}
		}
		return l_max;
	};

	//! function finishes up construction, inits all support structures etc.
	template<class xbwt_t>
	static void finish( sdsl::cache_config &config, sdsl::bit_vector &mr, 
	                    sdsl::bit_vector &P, xbwt_t &x ) {
		auto event = sdsl::memory_monitor::event("construction of L and Last");

		{
			//load bwt from cache
			text_type bwt;
			sdsl::load_from_cache(bwt, sdsl::conf::KEY_BWT, config);

			//  Compute L and Last by overwriting bwt and mr
			sdsl::bit_vector e(256, 0); // alphabet indicator
			size_type pos = 0;
			size_type count = 0;

			for (size_t i = 1; i < mr.size(); i++) {
				if (mr[i] && (i > 1)) { // new row range starts
					mr[pos - 1] = 1;      // Last overwriting MR
					// clear e for next row range
					for (size_t j = pos - count; j < pos; j++) {
						unsigned char c = bwt[j];
						e[c] = 0;
					}
					count = 0;
				}
				unsigned char c = bwt[i];
				if (c == 0) c = xbwt_t::delimchar; //treat nullbyte as delimiter
				if (e[c] == 0) { //character occurs for first time in row range
					mr[pos] = 0;  // Last overwriting MR
					bwt[pos] = c; // L overwriting BWT
					e[c] = 1;
					pos++;
					count++;
				}
			}
			mr[pos - 1] = 1;

			// cut arrays
			mr.resize(pos);
			bwt.resize(pos);

			//compute C array
			x.C.resize( 256 );
			sdsl::util::set_to_value( x.C, 0 );

			for (size_t i = 0; i < bwt.size(); i++)	{
				x.C[bwt[i]]++;
			}
			x.C[xbwt_t::delimchar] = 1;
			size_type sum = 1; //for empty string
			for (size_t i = xbwt_t::delimchar+1; i < x.C.size(); i++) {
				auto tmp = x.C[i];
				x.C[i] = sum;
				sum += tmp;
			}

			//store stuff to cache
			sdsl::store_to_cache( bwt, KEY_L, config); //store to save space
			sdsl::register_cache_file(KEY_L, config);
		}

		// construct wavelet tree for L
		text_buf_type l_buf( cache_file_name(KEY_L, config) );
		typename xbwt_t::wt_type L( l_buf, mr.size() );
		std::swap( L, x.L );

		// construct rank and select support for Last
		typename xbwt_t::bit_vector_type Last( std::move(mr) );
		std::swap( Last, x.Last );
		sdsl::util::init_support(x.Last_rank, &x.Last);
		sdsl::util::init_support(x.Last_select, &x.Last);

		//construct support for P
		std::swap( P, x.P );
		sdsl::util::init_support(x.P_bps_support, &x.P);
	};
};

template<xbwt_construct_algorithm a>
const char xbwt_construct<a>::KEY_MR[] = "mr";

template<xbwt_construct_algorithm a>
const char xbwt_construct<a>::KEY_RCP[] = "rcp";

template<xbwt_construct_algorithm a>
const char xbwt_construct<a>::KEY_L[] = "l";

//// MANZINI CONSTRUCTION /////////////////////////////////////////////////////

template<>
template<class xbwt_t>
void xbwt_construct<MAN>::construct( sdsl::cache_config &config, xbwt_t &x ) {
	size_type l_max = computeLmax<xbwt_t>( config );
	const char *KEY_TEXT = sdsl::key_text_trait<8>::KEY_TEXT;
	sdsl::bit_vector P;

	{
		// load text
		text_type text;
		sdsl::load_from_cache(text, KEY_TEXT, config);

		//create buffers to store rcp and mr
		sdsl::int_vector_buffer<> rcp_buf(sdsl::cache_file_name(KEY_RCP, config),
			                    std::ios::out, sdsl::bits::hi(l_max) + 1);
		sdsl::register_cache_file(KEY_RCP, config);
		sdsl::int_vector_buffer<1> mr_buf(sdsl::cache_file_name(KEY_MR, config), std::ios::out);
		sdsl::register_cache_file(KEY_MR, config);

		//compute rcp and mr
		{
			auto event = sdsl::memory_monitor::event("construction of MR and RCP");

			sdsl::int_vector_buffer<> sa_buf(sdsl::cache_file_name(sdsl::conf::KEY_SA, config));

			//	Calculate PHI (stored in array prcp)
			size_type t_len = sa_buf.size();
			sdsl::int_vector<> prcp(t_len, 0, sdsl::bits::hi(t_len) + 1);
			sdsl::bit_vector pmr(t_len, 1);
			for (size_type i = 0, sai_1 = 0; i < t_len; ++i) {
				size_type sai = sa_buf[i];
				prcp[sai] = sai_1;
				sai_1 = sai;
			}

			//  Calculate permuted RCP and MR array (text order), called PRCP and PMR
			for (size_type i = 0, l = 0; i < t_len - 1; ++i) {
				size_type phii = prcp[i];
				while (text[i + l] == text[phii + l]) {
					if (text[i + l] == xbwt_t::delimchar) {
						pmr[i] = 0;
						break;
					}
					++l;
				}
				prcp[i] = l;
				if (l)	--l;
			}

			//	Transform PRCP and PMR into RCP and MR
			rcp_buf[0] = 0; // handle 0-byte at first position
			mr_buf[0] = 0;  // handle 0-byte at first position
			rcp_buf[1] = 0;
			mr_buf[1] = 1;
			for (size_type i = 2; i < t_len; ++i) {
				size_type sai = sa_buf[i];
				rcp_buf[i] = prcp[sai];
				mr_buf[i] = pmr[sai];
			}
		}

		// construct P
		{
			auto event = sdsl::memory_monitor::event("construction of P");
			//set P to a good size
			P.resize( 2 * mr_buf.size() );
			sdsl::util::set_to_value( P, 0 );

			// load suffix array from cache
			sdsl::int_vector<> sa;
			sdsl::load_from_cache(sa, sdsl::conf::KEY_SA, config);

			std::stack<std::pair<size_type, size_type>> s; // (i, l)
			size_type l = 0;
			size_type pos = 0;

			s.push(std::make_pair(1, 0)); // ignore entry for 0-byte at first position
			P[pos] = 1;
			pos++;

			for (size_type i = 2; i < mr_buf.size(); i++) {
				if (mr_buf[i]) {
					l = rcp_buf[i] + 1;
					while (s.top().second >= l) {
						s.pop();
						pos++;
					}
					if (text[sa[s.top().first] + l - 1] != xbwt_t::delimchar) {
						l = s.top().second;
						s.pop();
						pos++;
					}
					s.push(std::make_pair(i, l));
					P[pos] = 1;
					pos++;
				}
			}
			while (!s.empty()) {
				s.pop();
				pos++;
			}
			P.resize(pos);
		}
	}
	//finish construction
	sdsl::bit_vector mr;
	sdsl::load_from_cache(mr, KEY_MR, config);
	finish( config, mr, P, x );
};

//// MANZINI LIGHTWEIGHT CONSTRUCTION /////////////////////////////////////////

template<>
template<class xbwt_t>
void xbwt_construct<MAN_LW>::construct( sdsl::cache_config &config, xbwt_t &x ) {
	typedef typename xbwt_t::wt_type t_wt_type;
	size_type l_max = computeLmax<xbwt_t>( config );
	sdsl::bit_vector mr;
	sdsl::bit_vector P;

	//build wavelet tree from bwt
	text_buf_type bwt_buf(sdsl::cache_file_name(sdsl::conf::KEY_BWT, config));
	t_wt_type wt(bwt_buf, bwt_buf.size());
	std::vector<uint64_t> C;
	sdsl::create_C_array(C, wt);

	//compute MR
	{
		auto event = sdsl::memory_monitor::event("construction of MR");

		mr.resize(wt.size());             // allocate mr
		sdsl::util::set_to_value(mr, 0);

		size_type b = 1; // ignore entry for 0-byte at postion 0
		size_type e = wt.rank(wt.size(), xbwt_t::delimchar) + 1;
		std::stack<std::pair<size_type, size_type>> s;
		s.push(std::make_pair(b, e));
		size_type k;
		std::vector<typename t_wt_type::value_type> cs(wt.sigma);
		std::vector<typename t_wt_type::size_type> rank_c_b(wt.sigma);
		std::vector<typename t_wt_type::size_type> rank_c_e(wt.sigma);

		while (!s.empty()) {
			b = s.top().first;
			e = s.top().second;
			s.pop();

			sdsl::interval_symbols(wt, b, e, k, cs, rank_c_b, rank_c_e);
			mr[b] = 1; // mark start position of maximal range
			for (size_type i = 0; i < k; i++) {
				auto c = cs[i];
				if (c != xbwt_t::delimchar && c != 0) {
					size_type next_b = C[c] + rank_c_b[i];
					size_type next_e = C[c] + rank_c_e[i];
					s.push(std::make_pair(next_b, next_e));
				}
			}
		}
	}

	//compute P
	{
		auto event = sdsl::memory_monitor::event("construction of P");
		// compute n = size of pi array
		sdsl::rank_support_v<> mr_rank(&mr);
		size_type n_ = mr_rank(mr.size());

		// compute rcp_ and len_ array
		sdsl::int_vector<> rcp_(n_ + 1, 0, sdsl::bits::hi(l_max + 1) + 1);
		sdsl::int_vector<> len_(n_, 0, sdsl::bits::hi(l_max) + 1);

		{
			std::queue<std::tuple<size_type, size_type, size_type, bool>> q;
			size_type n = wt.size();

			typename t_wt_type::size_type k;
			std::vector<typename t_wt_type::value_type> cs(wt.sigma);
			std::vector<typename t_wt_type::size_type> rank_c_b(wt.sigma);
			std::vector<typename t_wt_type::size_type> rank_c_e(wt.sigma);

			// note: rcp_ values are increased by 1 so that 0 can be treated as bottom
			len_[0] = 0;
			rcp_[0] = 1;
			rcp_[rcp_.size() - 1] = 1;

			// initialize queue
			sdsl::interval_symbols(wt, 0, n, k, cs, rank_c_b, rank_c_e);
			for (size_type i = 0; i < k; i++) {
				auto c = cs[i];
				bool x = false;
				if (c == xbwt_t::delimchar) { // mark intervalls to count len_
					x = true;
				}
				size_type lb = C[c] + rank_c_b[i];
				size_type rb = C[c] + rank_c_e[i];
				size_type pos = mr_rank(rb);
				if (rcp_[pos] == 0) {
					rcp_[pos] = 1;
					q.push(std::make_tuple(lb, rb, 2, x));
				}
			}

			// compute rcp_ and len_
			while (!q.empty()) {
				size_type b = std::get<0>(q.front());
				size_type e = std::get<1>(q.front());
				size_type l = std::get<2>(q.front());
				bool x = std::get<3>(q.front());
				q.pop();
				if (x)
					len_[mr_rank(b)] = l - 2; // write len_ if marked
				sdsl::interval_symbols(wt, b, e, k, cs, rank_c_b, rank_c_e);
				for (size_type i = 0; i < k; i++) {
					auto c = cs[i];
					if (c != xbwt_t::delimchar && c != 0) {
						size_type lb = C[c] + rank_c_b[i];
						size_type rb = C[c] + rank_c_e[i];
						size_type pos = mr_rank(rb);
						if (rcp_[pos] == 0) {
							rcp_[pos] = l;
							q.push(std::make_tuple(lb, rb, l + 1, x));
						} else if (x) { // make sure to calculate all len_ entrys
							q.push(std::make_tuple(lb, rb, l + 1, x));
						}
					}
				}
			}
		}

		//get rid of wavelet tree
		wt = t_wt_type();

		//finally, compute P
		P.resize( 2 * n_ );
		sdsl::util::set_to_value( P, 0 );
		std::stack<std::pair<size_type, size_type>> s;
		size_type l = 0;
		size_type pos = 0;

		// note: rcp_ values are already increased by 1
		s.push(std::make_pair(0, 0));
		P[pos++] = 1;

		for (size_type i = 1; i < n_; i++) {
			l = rcp_[i];
			while (s.top().second >= l) {
				s.pop();
				pos++;
			}
			if (len_[s.top().first] != l - 1) {
				l = s.top().second;
				s.pop();
				pos++;
			}
			s.push(std::make_pair(i, l));
			P[pos] = 1;
			pos++;
		}
	}
	//finish construction
	finish( config, mr, P, x );
};

//// OHLEBUSCH ////////////////////////////////////////////////////////////////

template<>
template<class xbwt_t>
void xbwt_construct<OHL>::construct( sdsl::cache_config &config, xbwt_t &x ) {
	typedef typename xbwt_t::wt_type t_wt_type;
	size_type l_max = computeLmax<xbwt_t>( config );
	sdsl::bit_vector mr;
	sdsl::bit_vector P;

	{
		sdsl::int_vector<> cc(0, 0, sdsl::bits::hi(l_max) + 1);
		// calculate MR and Cc array
		{
			auto event = sdsl::memory_monitor::event("construction of mr and cc");

			//build wavelet tree from bwt
			text_buf_type bwt_buf(sdsl::cache_file_name(sdsl::conf::KEY_BWT, config));
			t_wt_type wt(bwt_buf, bwt_buf.size());
			std::vector<uint64_t> C;
			sdsl::create_C_array(C, wt);

			mr.resize(wt.size());             // allocate mr and cc
			sdsl::util::set_to_value(mr, 0);
			cc.resize( wt.size() );
			sdsl::util::set_to_value(cc, 0);
		

			size_type b = 1;	// ignore entry for 0-byte at postion 0
			size_type e1 = wt.rank(wt.size(), xbwt_t::delimchar) + 1;
			size_type e2 = mr.size();
	  		std::stack<std::tuple<size_type, size_type, size_type>> s;
			s.push(std::make_tuple(b, e1, e2));

			typename t_wt_type::size_type k;
			std::vector<typename t_wt_type::value_type> cs(wt.sigma);
			std::vector<typename t_wt_type::size_type> rank_c_b(wt.sigma);
			std::vector<typename t_wt_type::size_type> rank_c_e1(wt.sigma);

			while (!s.empty()) {
				b = std::get<0>(s.top());
				e1 = std::get<1>(s.top());
				e2 = std::get<2>(s.top());
				s.pop();

				mr[b] = 1;    // mark start position of maximal range
				cc[e2 - 1]++; // count endpositions (not inclusive)

				sdsl::interval_symbols(wt, b, e1, k, cs, rank_c_b, rank_c_e1);

				for (size_type i = 0; i < k; i++) {
					auto c = cs[i];
					if (c != xbwt_t::delimchar && c != 0) {
						size_type next_b = C[c] + rank_c_b[i];
						size_type next_e1 = C[c] + rank_c_e1[i];
						size_type next_e2;
						if (e2 == e1) {
							next_e2 = next_e1;
						} else {
							next_e2 = C[c] + wt.rank(e2, c);
						}
						s.push(std::make_tuple(next_b, next_e1, next_e2));
					}
				}
			}
		}

		// calculate P
		{
			auto event = sdsl::memory_monitor::event("construction of P");
			size_type m = mr.size();
			P.resize(2 * m);

			size_type pos = 0;
			for (size_type i = 0; i < m; i++) {
				if (mr[i]) {
					P[pos++] = 1;
				}
				for (size_type j = 0; j < cc[i]; j++) {
					P[pos++] = 0;
				}
			}
			P.resize(pos);
		}
	}
	//finish construction
	finish( config, mr, P, x );
};

//// OHLEBUSCH LIGHTWEIGHT ////////////////////////////////////////////////////

template<>
template<class xbwt_t>
void xbwt_construct<OHL_LW>::construct( sdsl::cache_config &config, xbwt_t &x ) {
	typedef typename xbwt_t::wt_type t_wt_type;
	size_type l_max = computeLmax<xbwt_t>( config );
	sdsl::bit_vector mr;
	sdsl::bit_vector P;

	{
		//build wavelet tree from bwt
		text_buf_type bwt_buf(sdsl::cache_file_name(sdsl::conf::KEY_BWT, config));
		t_wt_type wt(bwt_buf, bwt_buf.size());
		std::vector<uint64_t> C;
		sdsl::create_C_array(C, wt);

		// calculate MR array
		{
			auto event = sdsl::memory_monitor::event("construction of mr");
			mr.resize(wt.size());             // allocate mr and cc
			sdsl::util::set_to_value(mr, 0);

			size_type b = 1; // ignore entry for 0-byte at postion 0
			size_type e = wt.rank(wt.size(), xbwt_t::delimchar) + 1;
			std::stack<std::pair<size_type, size_type>> s;
			s.push(std::make_pair(b, e));
			typename t_wt_type::size_type k;
			std::vector<typename t_wt_type::value_type> cs(wt.sigma);
			std::vector<typename t_wt_type::size_type> rank_c_b(wt.sigma);
			std::vector<typename t_wt_type::size_type> rank_c_e(wt.sigma);

			while (!s.empty()) {
				b = s.top().first;
				e = s.top().second;
				s.pop();

				sdsl::interval_symbols(wt, b, e, k, cs, rank_c_b, rank_c_e);
				mr[b] = 1; // mark start position of maximal range
				for (size_type i = 0; i < k; i++) {
					auto c = cs[i];
					if (c != xbwt_t::delimchar && c != 0) {
						size_type next_b = C[c] + rank_c_b[i];
						size_type next_e = C[c] + rank_c_e[i];
						s.push(std::make_pair(next_b, next_e));
					}
				}
			}
		}

		//calculate cc array
		sdsl::int_vector<> cc_(0, 0, sdsl::bits::hi(l_max) + 1);
		{
			auto event = sdsl::memory_monitor::event("construction of cc");
			sdsl::rank_support_v<> mr_rank(&mr);
			size_type n = mr_rank(mr.size());
			cc_.resize( n );
			sdsl::util::set_to_value( cc_, 0 );

			size_type b = 1;
			size_type e1 = wt.rank(wt.size(), xbwt_t::delimchar) + 1;
			size_type e2 = mr.size();

			std::stack<std::tuple<size_type, size_type, size_type>> s;
			s.push(std::make_tuple(b, e1, e2));

			typename t_wt_type::size_type k;
			std::vector<typename t_wt_type::value_type> cs(wt.sigma);
			std::vector<typename t_wt_type::size_type> rank_c_b(wt.sigma);
			std::vector<typename t_wt_type::size_type> rank_c_e1(wt.sigma);

			while (!s.empty()) {
				b = std::get<0>(s.top());
				e1 = std::get<1>(s.top());
				e2 = std::get<2>(s.top());
				s.pop();

				cc_[mr_rank(e2) - 1]++; // count endpositions (not inclusive)

				sdsl::interval_symbols(wt, b, e1, k, cs, rank_c_b, rank_c_e1);

				for (size_type i = 0; i < k; i++) {
					auto c = cs[i];
					if (c != xbwt_t::delimchar && c != 0) {
						size_type next_b = C[c] + rank_c_b[i];
						size_type next_e1 = C[c] + rank_c_e1[i];
						size_type next_e2;
						if (e2 == e1) {
							next_e2 = next_e1;
						} else {
							next_e2 = C[c] + wt.rank(e2, c);
						}
						s.push(std::make_tuple(next_b, next_e1, next_e2));
					}
				}
			}
		}

		//get rid of wavelet tree
		wt = t_wt_type();

		// calculate P
		{
			auto event = sdsl::memory_monitor::event("construction of P");
			size_type n = cc_.size();
			P.resize(2 * n);

			size_type k = 0;
			for (size_type i = 0; i < n; i++) {
				P[k] = 1;
				k++;
				for (size_type j = 0; j < cc_[i]; j++) {
					P[k] = 0;
					k++;
				}
			}
		}
	}
	//finish construction
	finish( config, mr, P, x );
};

#endif
