/*
 * mpsearch.cpp for trickier xbwt tricks
 * Copyright (c) 2018 Stefan Stauß, Uwe Baier All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include <algorithm>
#include <iostream>
#include <string>
#include <string.h>

using namespace std;

#include "xbwt.hpp"

using namespace std;
using namespace sdsl;

//TODO: change if wanted
typedef xbwt<> xbwt_t;

void printUsage( char **argv ) {
	cerr << "USAGE: " << argv[0] << " XBWTTRIEFILE | " << argv[0] << " PATTERN PATTERN ..." << endl;
	cerr << "DESCRIPTION: Program searches for all occurences of all listed patterns using stdin" << endl;
	cerr << "  If only 1 argument is given, it is interpreted as a serialized xbwt trie used for search" << endl;
	cerr << "  , output is a list of the endposition of an occurence as well as the lexicographic " << endl;
	cerr << "  rank of the reversed string." << endl << endl;
	cerr << "  If more than one arguments are given, program builds a trie of the given patterns" << endl;
	cerr << "  and outputs for each occurence a pair of the start position and the found string" << endl;
	cerr << "XBWTTRIEFILE: a file containing a serialized XBWT trie used for muli-pattern search" << endl;
	cerr << "PATTERN: pattern to be used for trie construction" << endl;
	cerr << "NOTE: to work correct, trie should not contain strings which are proper substrings" << endl;
	cerr << "  of other contained strings" << endl;
};

//aho-corasick-algorithm
template<class Output>
void ahocorasick( istream &in, const xbwt_t &x, Output output ) {
	auto v = x.root();
	unsigned char c = in.get();
	size_t i = 0;
	while (in) {
		auto v_ = x.get_child( v, c );
		if (v_ != x.root()) { //follow edge
			auto llr = x.get_child( v_, xbwt_t::delimchar ); //leaf lex rank
			if (llr != x.root()) { //v_ is a leaf, so we detected string
				output( x, v_, i, llr );
			}
			v = v_;
			c = in.get();
			i++;
		}
		else if (v == x.root()) {
			c = in.get();
			i++;
		}
		else {
			v = x.suffix_link( v );
		}
	}
};

int main( int argc, char **argv ) {
	xbwt_t x;
	//check arguments
	if (argc < 2) {
		printUsage( argv );
		return 1;
	}
	//// MODE 1: load serialized trie ////
	if (argc == 2) {
		//load serialized trie from file
		load_from_file( x, argv[1] );
		//run aho-corasick algorithm using normal output
		ahocorasick( cin, x, []( const xbwt_t &x, xbwt_t::size_type v, size_t i, xbwt_t::size_type leaflex ) {
			cout << (i + 1) << "\t" << leaflex << endl; } );
	}
	//// MODE 2: create trie from arguments ////	
	else {
		//generate string from which trie can be built
		size_t lensum = 1; //for nullbyte
		for (int i = 1; i < argc; i++) {
			size_t l = strlen( argv[i] );
			if (l == 0) {
				printUsage( argv );
				cerr << "No empty patterns allowed" << endl;
				return 1;
			}
			//update lensum and reverse argument
			lensum += l+1;
			reverse( argv[i], argv[i] + l );
		}
		//build up string
		char *str = new char[lensum];
		char *s = str;
		for (int i = 1; i < argc; i++) {
			strcpy(s, argv[i] ); //copy next (reversed) string
			s = strchr(s, '\0'); //go to last position in string
			*(s++) = xbwt_t::delimchar; //insert delimiter
		}
		*s = '\0';
		//construct trie using some settings
		construct_config::byte_algo_sa = LIBDIVSUFSORT; //fastest algorithms
		xbwt_t::construct_algorithm = MAN;
		construct_im( x, (const char *)str, 1 );
		delete[] str;

		//run aho-corasick algorithm using excessive output
		ahocorasick( cin, x, []( const xbwt_t &x, xbwt_t::size_type v, size_t i, xbwt_t::size_type leaflex ) {
			string s;
			while (v != x.root()) {
				s.push_back( x.get_parent_edge(v) );
				v = x.get_parent(v);
			}
			reverse( s.begin(), s.end() );
			cout << (i + 1 - s.size()) << ":\t" << s << endl; } );
	}
	return 0;
}	
