/*
 * trie_construct.cpp for trickier xbwt tricks
 * Copyright (c) 2018 Stefan Stauß, Uwe Baier All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <iostream>
#include <string>
#include <string.h>
#include <sdsl/bit_vectors.hpp>
#include <sdsl/wavelet_trees.hpp>

#include "xbwt.hpp"

using namespace std;
using namespace sdsl;

//TODO: change if wanted
typedef xbwt<> xbwt_t;
//typedef xbwt<wt_huff<rrr_vector<>>> xbwt_t;

void printUsage( char **argv ) {
	cerr << "USAGE: " << argv[0] << " [OPTIONS] TRIEINFILE XBWTTRIEFILE" << endl;
	cerr << "OPTIONS:" << endl;
	cerr << "  -i\tEnable informative mode, printing memory peaks (in bytes) and" << endl;
	cerr << "    \tconstruction timings (in milliseconds) during construction" << endl;
	cerr << "  -sa\tChoose suffix array construction algorithm. Must be followed by one of:" << endl;
	cerr << "     \tDIVSUFSORT   use divsufsort (fast but memory-intensive)" << endl;
	cerr << "     \tSE_SAIS         use a semi-external algorithm (slower but lower mem peak)" << endl;
	cerr << "  -xa\tChoose xbwt construction algorithm. Must be followed by one of:" << endl;
	cerr << "     \tMAN     use algorithm by Manzini" << endl;
	cerr << "     \tMAN_LW  use lightweight algorithm by Manzini" << endl;
	cerr << "     \tOHL     use algorithm by Ohlebusch et al." << endl;
	cerr << "     \tOHL_LW  use lightweight algorithm by Ohlebusch et al." << endl;
	cerr << "TRIEINFILE:" << endl;
	cerr << "  File containing the strings from which the trie should be built from" << endl;
	cerr << "  Strings must be reversed and seperated by a '\\1' - byte ('\\1'-byte also at the end of a file), nullbytes are permitted" << endl;
	cerr << "XBWTTRIEFILE:" << endl;
	cerr << "  File where to store the serialized trie" << endl;
};

//little hack to get extra information from memory managemant
const format_type leet_format = (format_type)1337;
namespace sdsl {
	template<>
	void write_mem_log<leet_format>(ostream& out,const memory_monitor& m) {

		//get all memory events
		auto events = m.completed_events;
		std::sort( events.begin(), events.end() );
		auto e = events.begin();

		//scan events and detect peak and time of both suffix array and xbwt construction

		//scan for suffix array construction
		int64_t sa_mem_peak = 0;
		auto sa_start = m.start_log;
		auto sa_end = sa_start;		
		while (e != events.end() && e->name != "XBWT") {
			for (auto alloc : e->allocations) {
				sa_mem_peak = std::max(sa_mem_peak, alloc.usage);
				sa_end = alloc.timestamp;
			}
			++e;
		}

		//scan for xbwt construction
		int64_t xbwt_mem_peak = 0;
		auto xbwt_start = sa_end;
		auto xbwt_end = xbwt_start;
		while (e != events.end()) {
			for (auto alloc : e->allocations) {
				xbwt_mem_peak = std::max(xbwt_mem_peak, alloc.usage);
				xbwt_end = alloc.timestamp;
			}
			++e;
		}

		//print results
		out << "sa_mem_peak\t" << sa_mem_peak << endl;
		out << "sa_time\t" << chrono::duration_cast<chrono::milliseconds>(sa_end-sa_start).count() << endl;

		out << "xbwt_mem_peak\t" << xbwt_mem_peak << endl;
		out << "xbwt_time\t" << chrono::duration_cast<chrono::milliseconds>(xbwt_end-xbwt_start).count() << endl;
	};
};

int main( int argc, char **argv ) {
	//set default configuration
	construct_config::byte_algo_sa = SE_SAIS;
	xbwt_t::construct_algorithm = OHL;
	bool informative = false; //informative mode
	string infile = "data/dna.50MB";
	string outfile = "test.xbwt";

	//check parameters
	if (argc < 3) {
		printUsage( argv );
		cerr << "At least 2 parameters expected" << endl;
		return 1;
	}
	enum { SA, XA, IN, NO } last_option; //enumeration for last option
	last_option = NO;
	for (int i = 1; i < argc - 2; i++) { //analyze options
		switch (last_option) {
		case NO:
		case IN: //last option does not require a parameter, read next option
			if (strcmp(argv[i], "-i") == 0) { //enable informative mode
				last_option = IN;
				informative = true;
			}
			else if (strcmp(argv[i], "-sa") == 0) {
				last_option = SA;
			}
			else if (strcmp(argv[i], "-xa") == 0) {
				last_option = XA;
			}
			else {
				printUsage(argv);
				cerr << "Unknown option " << argv[i] << endl;
				return 1;
			}
			break;
		case SA: //choose suffix array construction algorithm
			if (strcmp(argv[i], "DIVSUFSORT") == 0) {
				construct_config::byte_algo_sa = LIBDIVSUFSORT;
			}
			else if (strcmp(argv[i], "SE_SAIS") == 0) {
				construct_config::byte_algo_sa = SE_SAIS;
			}
			else {
				printUsage( argv );
				cerr << "Unknown suffix array construction algorithm " << argv[i] << endl;
				return 1;
			}
			last_option = NO;
			break;
		case XA: //choose xbwt construction algorithm
			if (strcmp(argv[i], "MAN") == 0) {
				xbwt_t::construct_algorithm = MAN;
			}
			else if (strcmp(argv[i], "MAN_LW") == 0) {
				xbwt_t::construct_algorithm = MAN_LW;
			}
			else if (strcmp(argv[i], "OHL") == 0) {
				xbwt_t::construct_algorithm = OHL;
			}
			else if (strcmp(argv[i], "OHL_LW") == 0) {
				xbwt_t::construct_algorithm = OHL_LW;
			}
			else {
				printUsage( argv );
				cerr << "Unknown xbwt construction algorithm " << argv[i] << endl;
				return 1;
			}
			last_option = NO;
			break;
		}
	}
	infile = argv[argc-2];
	outfile = argv[argc-1];

	//construct xbwt
	xbwt_t x;
	memory_monitor::start();
	construct(x, infile, 1);
	memory_monitor::stop();
	
	//print infos if wanted
	if (informative) {
		memory_monitor::write_memory_log<leet_format>(cout);
	}

	//serialize result
	store_to_file(x, outfile);
	return 0;
}
